/* GM70 SE amplifier control module firmware
//   V2.1    12.10.2015
//   Vincent Duparc
//   not for commercial use.
*/

#define DEBUG

///////////////////////////////////
// parameters
///////////////////////////////////

// warm up delay before applying B+ (seconds)
const unsigned long WARMUP_DELAY = 15;

#define RES 10 //value of the cathode resistor. 

#define COEF RES * 204.6 //Arduino AnalogRead units

// max acceptable average current over one second:
const float MAX_AVG_I = 0.125 * COEF; //125mA

// max instantaneous current:
const float MAX_I = 0.25 * COEF; //250mA

// clipping indicator
// (0.18A for 5500ohm load, 0.14A for 9000ohm)
const float SIGNAL_PEAK = 0.14 * COEF;

// transformer max temperature (°C * 2)
const int MAX_TEMP = 70 * 2;

// LED warmup timing (msec)
const int TIME_LED_WUP_WUP_ON = 2050;
const int TIME_LED_WUP_WUP_OFF = 150;
const int TIME_LED_WUP_WUP_PAUSE = 800;
// LED overload timing (msec)
const int TIME_LED_WUP_OVL_ON = 100;
const int TIME_LED_WUP_OVL_OFF = 100;
const int TIME_LED_WUP_OVL_PAUSE = 1000;
// LED overheat timing (msec)
const int TIME_LED_WUP_TEMP_ON = 100;
const int TIME_LED_WUP_TEMP_OFF = 150;


// LED intensity
const int PWM_LED_MAX = 250; // LED max intensity, 0..255
const int PWM_LED_RUN = 50;  // LED intensity in normal status


// onboard LED timing (msec)
const int BRD_LED_WUP_ON = 500;
const int BRD_LED_WUP_OFF = 500;
const int BRD_LED_RUN_ON = 30;
const int BRD_LED_RUN_OFF = 970;

// inrush limiter timing (msec)
const int INRUSH_DELAY = 2000;

// I/O pins
const int PIN_LED_WUP = 3;       // warmup/run LED pin
const int PIN_LED_BRD = 13;      // onboard LED pin
const int PIN_REL_POWER = 9;     // power relay pin
const int PIN_REL_INRUSH1 = 7;   // inrush relay pin
const int PIN_REL_INRUSH2 = 5;   // inrush relay pin
const int PIN_IN_LEFT = A1;      // analog in left tube cathode pin
const int PIN_IN_RIGHT = A0;     // analog in right tube cathode pin
const int PIN_TEMP_1 = A6;       // temp sensor 1
const int PIN_TEMP_2 = A7;	 // temp sensor 2

//----------------------------------------------------------
// status levels.
const int STATUS_WARMUP = 1;
const int STATUS_RUN = 2;
const int STATUS_ERR_OVL_L = 3;
const int STATUS_ERR_OVL_R = 4;
const int STATUS_ERR_TEMP = 5;


// global variables
unsigned long mil_0;      // ticks at t zero
unsigned long mil_now;    // ticks now
unsigned long sec_0;
unsigned long sum_I_L;
unsigned long sum_I_R;
unsigned long samples;
int status;               // global status
bool onPeak;


///////////////////////////////////
// SETUP
///////////////////////////////////
void setup() {
  //port setup
  for (int i = 0; i <= 13; i++) {
    pinMode(i, INPUT);
    digitalWrite(i, LOW);
  }
  pinMode(PIN_LED_WUP, OUTPUT);
  pinMode(PIN_LED_BRD, OUTPUT);
  pinMode(PIN_REL_POWER, OUTPUT);
  pinMode(PIN_REL_INRUSH1, OUTPUT);
  pinMode(PIN_REL_INRUSH2, OUTPUT);
  pinMode(PIN_IN_LEFT, INPUT);
  pinMode(PIN_IN_RIGHT, INPUT);
  pinMode(PIN_TEMP_1, INPUT);
  pinMode(PIN_TEMP_2, INPUT);

#ifdef DEBUG
  Serial.begin(57600);
  Serial.println("\nHello!\n");
#endif


  //go!
  digitalWrite(PIN_LED_WUP, HIGH);
  delay(50);
  digitalWrite(PIN_LED_WUP, LOW);
  delay(50);
  digitalWrite(PIN_LED_WUP, HIGH);
  delay(50);
  digitalWrite(PIN_LED_WUP, LOW);

  delay(500); // plus Arduino bootup time
  digitalWrite(PIN_REL_INRUSH2, HIGH);
  status = STATUS_WARMUP;
  mil_0 = millis();
#ifdef DEBUG
  Serial.println("Warmup...");
#endif

  if (analogRead(PIN_IN_LEFT) + analogRead(PIN_IN_RIGHT) > 10) {
    status = STATUS_ERR_TEMP;
#ifdef DEBUG
    Serial.println("Not connected to cathode"); //reading abnormal current values - probably inputs are floating
#endif
  }

}



///////////////////////////////////
// MAIN
///////////////////////////////////
void loop() {

  mil_now = millis() - mil_0;

  /////////////////
  // STATUS_WARMUP
  /////////////////
  if (status == STATUS_WARMUP) {

    //onboard LED
    if (mil_now % (BRD_LED_WUP_ON + BRD_LED_WUP_OFF) <= BRD_LED_WUP_ON)
      digitalWrite(PIN_LED_BRD, HIGH);
    else
      digitalWrite(PIN_LED_BRD, LOW);


    //warm-up LED
    unsigned long wup_cycle_ticks = mil_now % (TIME_LED_WUP_WUP_ON + TIME_LED_WUP_WUP_OFF + TIME_LED_WUP_WUP_PAUSE);
    byte pw = sinMap(wup_cycle_ticks, 0, TIME_LED_WUP_WUP_ON);
    boolean isOn = false;

    if (wup_cycle_ticks < TIME_LED_WUP_WUP_ON) {
      analogWrite(PIN_LED_WUP, pw);
      isOn = true;
    }
    else if (wup_cycle_ticks < TIME_LED_WUP_WUP_ON + TIME_LED_WUP_WUP_OFF) {
      analogWrite(PIN_LED_WUP, map(wup_cycle_ticks, TIME_LED_WUP_WUP_ON, TIME_LED_WUP_WUP_ON + TIME_LED_WUP_WUP_OFF, PWM_LED_MAX, 0));
      isOn = true;
    }
    else {
      analogWrite(PIN_LED_WUP, 0);
      isOn = false;
    }


    // POWER RELAY
    if (mil_now >= (mil_0 + WARMUP_DELAY * 1000L) && !isOn) {
#ifdef DEBUG
      Serial.println("POWER UP!");
#endif
      analogWrite(PIN_LED_WUP, PWM_LED_RUN);
      digitalWrite(PIN_REL_POWER, HIGH);
      delay(INRUSH_DELAY);
      digitalWrite(PIN_REL_INRUSH1, HIGH);
      sum_I_L = 0;
      sum_I_R = 0;
      samples = 1;
      status = STATUS_RUN;
      sec_0 = 0;
      mil_0 = millis();
      mil_now = millis() - mil_0;
    }
  }


  ////////////////
  // STATUS_RUN
  ////////////////
  if (status == STATUS_RUN) {

    // measurement:
    unsigned int i_L = analogRead(PIN_IN_LEFT);
    unsigned int i_R = analogRead(PIN_IN_RIGHT);
    unsigned int temp1 = analogRead(PIN_TEMP_1);
    unsigned int temp2 = analogRead(PIN_TEMP_2);


    // onboard LED
    if (mil_now % (BRD_LED_RUN_ON + BRD_LED_RUN_OFF) <= BRD_LED_RUN_ON)
      digitalWrite(PIN_LED_BRD, HIGH);
    else
      digitalWrite(PIN_LED_BRD, LOW);


    // clipping indicator
    if (i_L > SIGNAL_PEAK || i_R > SIGNAL_PEAK) {
      digitalWrite(PIN_LED_WUP, HIGH);
      onPeak = true;
    }
    else if (onPeak) {
      analogWrite(PIN_LED_WUP, PWM_LED_RUN);
      onPeak = false;
    }



    unsigned long sec_now = mil_now / 1000;

    if (sec_now != sec_0) { //once each second
#ifdef DEBUG
      Serial.print("Average I L: ");
      Serial.print(sum_I_L / samples);
      Serial.print(" R: ");
      Serial.print(sum_I_R / samples);
      Serial.print("   Temperature 1: ");
      Serial.print(temp1 / 2, DEC);
      Serial.print("C  2: ");
      Serial.print(temp2 / 2, DEC);
      Serial.println("C");
#endif

      // Average I
      if (sum_I_L / samples >= MAX_AVG_I) {
        shutdown();
        status = STATUS_ERR_OVL_L;
#ifdef DEBUG
        Serial.print("overload average LEFT: ");
        Serial.println(sum_I_L / samples);
#endif
      }
      else if (sum_I_R / samples >= MAX_AVG_I) {
        shutdown();
        status = STATUS_ERR_OVL_R;
#ifdef DEBUG
        Serial.print("overload average RIGHT: ");
        Serial.println(sum_I_R / samples);
#endif
      }

      // Temperature
      if (temp1 >= MAX_TEMP || temp2 >= MAX_TEMP) {
        shutdown();
        status = STATUS_ERR_TEMP;
#ifdef DEBUG
        Serial.print("overheat: ");
        Serial.print(temp1 / 2, DEC);
        Serial.print("C  2: ");
        Serial.print(temp2 / 2, DEC);
        Serial.println("C");
#endif
      }


      sec_0 = sec_now;
      sum_I_L = i_L;
      sum_I_R = i_R;
      samples = 1;
    }
    else {
      samples++;
      sum_I_L += i_L;
      sum_I_R += i_R;
    }



    // instantaneous I
    if (i_L >= MAX_I) {
      // maximum instant left I exceeded
      shutdown();
      status = STATUS_ERR_OVL_L;
#ifdef DEBUG
      Serial.print("overload instant LEFT: ");
      Serial.println(i_L);
#endif
    }
    else if (i_R >= MAX_I) {
      // maximum instant right I exceeded
      shutdown();
      status = STATUS_ERR_OVL_R;
#ifdef DEBUG
      Serial.print("overload instant RIGHT: ");
      Serial.println(i_R);
#endif
    }
  } // end if STATUS_RUN


  ////////////////
  // OVERLOAD
  ////////////////
  else if (status == STATUS_ERR_OVL_L || status == STATUS_ERR_OVL_R) {

    // flash LED to indicate faulty tube L/R
    // left
    if (status == STATUS_ERR_OVL_L) {
      unsigned long cycle_duration = TIME_LED_WUP_OVL_ON + TIME_LED_WUP_OVL_PAUSE;
      unsigned int cycle_mil = mil_now % cycle_duration;
      if (cycle_mil <= TIME_LED_WUP_OVL_ON) {
        digitalWrite(PIN_LED_WUP, HIGH);
        digitalWrite(PIN_LED_BRD, HIGH);
      }
      else {
        digitalWrite(PIN_LED_WUP, LOW);
        digitalWrite(PIN_LED_BRD, LOW);
      }
    }

    // right, flash twice
    else {
      unsigned long cycle_duration = TIME_LED_WUP_OVL_ON + TIME_LED_WUP_OVL_OFF +
                                     TIME_LED_WUP_OVL_ON + TIME_LED_WUP_OVL_PAUSE;
      unsigned int cycle_mil = mil_now % cycle_duration;
      if (cycle_mil <= TIME_LED_WUP_OVL_ON) {
        digitalWrite(PIN_LED_WUP, HIGH);
        digitalWrite(PIN_LED_BRD, HIGH);
      }
      else if (cycle_mil <= TIME_LED_WUP_OVL_ON + TIME_LED_WUP_OVL_OFF) {
        digitalWrite(PIN_LED_WUP, LOW);
        digitalWrite(PIN_LED_BRD, LOW);
      }
      else if (cycle_mil <= TIME_LED_WUP_OVL_ON + TIME_LED_WUP_OVL_OFF + TIME_LED_WUP_OVL_ON) {
        digitalWrite(PIN_LED_WUP, HIGH);
        digitalWrite(PIN_LED_BRD, HIGH);
      }
      else {
        digitalWrite(PIN_LED_WUP, LOW);
        digitalWrite(PIN_LED_BRD, LOW);
      }
    }
  }

  ///////////
  // OVERHEAT
  else if (status == STATUS_ERR_TEMP) {

    if (mil_now % (TIME_LED_WUP_TEMP_ON + TIME_LED_WUP_TEMP_OFF) <= TIME_LED_WUP_TEMP_ON)
    {
      digitalWrite(PIN_LED_WUP, HIGH);
      digitalWrite(PIN_LED_BRD, HIGH);
    }
    else
    {
      digitalWrite(PIN_LED_WUP, LOW);
      digitalWrite(PIN_LED_BRD, LOW);
    }
  }

}




///////////////////////////////////
void shutdown() {
  digitalWrite(PIN_REL_POWER, LOW);
  digitalWrite(PIN_REL_INRUSH1, LOW);

#ifdef DEBUG
  Serial.println("*SHUTDOWN*");
#endif
}


///////////////////////////////////
byte sinMap(unsigned long x, int fr1, int fr2) {
  float coef = (sin(map(x, fr1, fr2, 4712, 6283) / 1000.0) + 1) * PWM_LED_MAX;
  return coef;
}
